#############################################################################
## Various file-related utilities.
##
## --------------------------------------------------------------------------
## Copyright (C) 2008 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: fileutils.rb,v 1.8 2013/05/03 17:52:52 youngh Exp $
#############################################################################

module ArkUtil

  # Returns a "handle" to a lock file opened with exclusive access, or nil on
  # error.  In some uncommon error scenarios, this method will raise an
  # exception (e.g., if a lock file already exists and is owned by someone
  # else).
  #
  # You should release the lock eventually by invoking release_exclusive_lock().
  # However, if your process dies while holding a lock, then the lock is
  # effectively released.
  def self.obtain_exclusive_lock(path)
    retval = nil
    num_tries = 0
    begin
      file_exists = false
      begin
        retval = File.open path, File::CREAT|File::EXCL|File::WRONLY, 0600
      rescue Errno::EEXIST
        file_exists = true
      end
      retval = File.open path, File::WRONLY, 0600 if file_exists
    rescue Errno::ENOENT  # handle a very rare race condition
      num_tries += 1
      raise if num_tries == 3
      retry 
    end

    unless retval
      $stderr.puts "ERROR: couldn't create/open lock file '#{path}'"
      return nil
    end

    if retval.flock File::LOCK_EX|File::LOCK_NB
      retval.truncate 0
      retval.puts $$
      retval.fsync
      retval.instance_variable_set :@___saved_lock_file_path, path
      return retval
    else
      retval.close
      $stderr.puts "ERROR: couldn't obtain exclusive lock on file '#{path}'; perhaps another process is holding the lock"
      return nil
    end
  end


  def self.release_exclusive_lock(lock)
    path = lock.instance_variable_get :@___saved_lock_file_path
    if path
      File.delete path
      lock.flock File::LOCK_UN
      lock.close
    end
  end


  def self.compute_sha1(path)
    compute_openssl_hash path, "sha1", "SHA1"
  end


  def self.compute_md5(path)
    compute_openssl_hash path, "md5", "MD5"
  end


  # Not really meant to be called from outside ArkUtil.
  def self.compute_openssl_hash(path, subcommand, hash_name)
    retval = nil
    ENV['ARKUTIL_COMPUTE_HASH_PATH'] = path
    command = "openssl #{subcommand} \"$ARKUTIL_COMPUTE_HASH_PATH\""
    IO.popen(command) do |pipe|
      pipe.each_line do |line|
        retval = $1 if line =~ /^#{hash_name}.*?= ([a-z0-9]+)$/
      end

      unless retval
        command = "openssl #{subcommand} #{path}"
        abort "ERROR: no hash value found in output of '#{command}'"
      end
    end
      
    unless $?.success?
      abort "ERROR: couldn't execute: #{command}"
    end  

    retval
  end

end  # module ArkUtil

