#############################################################################
## Various utilities for handling IP addresses.
##
## --------------------------------------------------------------------------
## Copyright (C) 2007 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: iputils.rb,v 1.11 2016/04/21 03:17:08 youngh Exp $
#############################################################################

require 'ipaddr'

module ArkUtil

class IPUtils

  # Index with 0 <= length <= 32
  # verify with: (0..32).each {|i| puts IPUtils::PREFIX_NET_MASK[i].to_s(2)}
  PREFIX_NET_MASK = [ 0x00000000, 0x80000000, 0xC0000000, 0xE0000000,
                      0xF0000000, 0xF8000000, 0xFC000000, 0xFE000000,
                      0xFF000000, 0xFF800000, 0xFFC00000, 0xFFE00000,
                      0xFFF00000, 0xFFF80000, 0xFFFC0000, 0xFFFE0000,
                      0xFFFF0000, 0xFFFF8000, 0xFFFFC000, 0xFFFFE000,
                      0xFFFFF000, 0xFFFFF800, 0xFFFFFC00, 0xFFFFFE00,
                      0xFFFFFF00, 0xFFFFFF80, 0xFFFFFFC0, 0xFFFFFFE0,
                      0xFFFFFFF0, 0xFFFFFFF8, 0xFFFFFFFC, 0xFFFFFFFE,
                      0xFFFFFFFF ]

  # Index with 0 <= length <= 32
  # verify with: (0..32).each {|i| puts IPUtils::PREFIX_HOST_MASK[i].to_s(2)}
  PREFIX_HOST_MASK = [ 0xFFFFFFFF, 0x7FFFFFFF, 0x3FFFFFFF, 0x1FFFFFFF,
                       0x0FFFFFFF, 0x07FFFFFF, 0x03FFFFFF, 0x01FFFFFF,
                       0x00FFFFFF, 0x007FFFFF, 0x003FFFFF, 0x001FFFFF,
                       0x000FFFFF, 0x0007FFFF, 0x0003FFFF, 0x0001FFFF,
                       0x0000FFFF, 0x00007FFF, 0x00003FFF, 0x00001FFF,
                       0x00000FFF, 0x000007FF, 0x000003FF, 0x000001FF,
                       0x000000FF, 0x0000007F, 0x0000003F, 0x0000001F,
                       0x0000000F, 0x00000007, 0x00000003, 0x00000001,
                       0x00000000 ]

  # {addr} should be an integer giving an IP address in host byte order.
  # Returns dotted decimal string.
  def self.addr_to_s(addr)
    retval = ""
    4.times do
      retval = (addr & 0xFF).to_s + "." + retval
      addr >>= 8
    end
    retval.chop!
  end

  # {addr} should be an IP address in dotted decimal format.
  # Returns integer, or nil on error.
  def self.parse_addr(addr)
    if addr =~ /^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})$/
      retval = 0
      for n in [$1, $2, $3, $4]
	n = n.to_i
	return nil if n > 255
	retval <<= 8
	retval |= n
      end
      retval
    else
      nil
    end
  end

  # {prefix} should be a string of the form "addr/length".
  # Returns [ addr, length ], or nil on error.
  def self.parse_prefix(prefix)
    addr, length = prefix.split "/", 2
    return nil if length == nil

    return nil unless length =~ /^\d+$/;
    length = length.to_i
    return nil unless length >= 0 && length <= 32

    addr = parse_addr addr
    if addr
      [ addr & PREFIX_NET_MASK[length], length ]
    else
      nil
    end
  end

  # {prefix} should be a string of the form "addr/length".
  # Returns [ start_addr, end_addr ], or nil on error.
  def self.parse_prefix_to_range(prefix)
    addr, length = prefix.split "/", 2
    return nil if length == nil

    return nil unless length =~ /^\d+$/;
    length = length.to_i
    return nil unless length >= 0 && length <= 32

    addr = parse_addr addr
    if addr
      [ addr & PREFIX_NET_MASK[length], addr | PREFIX_HOST_MASK[length] ]
    else
      nil
    end
  end

  # {addr} must be an integer IP address, not dotted-decimal formatted string.
  # Returns [ start_addr, end_addr ], or nil on error.
  def self.prefix_to_range(addr, length)
    return nil unless length >= 0 && length <= 32
    start_addr = addr & PREFIX_NET_MASK[length]
    end_addr = addr | PREFIX_HOST_MASK[length]
    [ start_addr, end_addr ]
  end

  # {addr} must be an integer IP address, not dotted-decimal formatted string.
  # Returns network portion of {addr} as integer value.
  def self.network_part(addr, length)
    return nil unless length >= 0 && length <= 32
    addr & PREFIX_NET_MASK[length]
  end

  # {addr} must be an integer IP address, not dotted-decimal formatted string.
  # Returns host portion of {addr} as integer value.
  def self.host_part(addr, length)
    return nil unless length >= 0 && length <= 32
    addr & PREFIX_HOST_MASK[length]
  end


  # -------------------------------------------------------------------------

  # {addr} must be an ASCII (dotted decimal formatted) IPv4 address.
  # Returns the destination as an ASCII address.
  def self.generate_random_destination(prefix, length)
    addr = IPAddr.new(prefix).to_i
    destination = generate_random_destination_aux addr, length
    IPAddr.new_ntoh(pack_iton(destination)).to_s
  end

  # {prefix} must be an integer IPv4 address, not a dotted decimal formatted
  # string.  All of the host bits of {prefix} must be zero.
  # Returns the destination as an integer.
  def self.generate_random_destination_aux(prefix, length)
    fail ArgumentError if prefix < 0 || length < 0 || length > 32
    prefix | rand(2 ** (32 - length))
  end

  # {addr} must be an integer IPv4 address, not a dotted decimal string.
  # Returns the IP address packed into a string in network byte order.
  def self.pack_iton(addr)
    fail ArgumentError if addr < 0
    return [addr].pack("N")
  end
end

end  # module ArkUtil
