#############################################################################
## A high-level class for asynchronously interacting with Scamper over its
## control socket.  Use this to issue traceroute commands to scamper and
## to fetch warts results over the control socket.
##
## --------------------------------------------------------------------------
## Copyright (C) 2008-2013 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: scamperdatafeed.rb,v 1.43 2020/03/26 06:20:59 youngh Exp $
#############################################################################

require 'socket'
require 'fcntl'
require 'tempfile'

require 'wartslib'

module ArkUtil

class ScamperDataFeed

  class StderrPrinter
    def method_missing(name, *args)
      args[0] = name.to_s.upcase! + ": " + args[0] + "\n"
      $stderr.printf *args
    end
  end

  WARTS_LIST_OBJECT = 'A$@4``0```!D````!`````#$R-RXP+C`N,3HV,SDT-P``'.unpack("u").first
  WARTS_CYCLE_OBJECT = 'M$@4``@```"4````!`````0````%1Q+A5`@`2;FEB8FQE<BYC86ED82YO<F<`'.unpack("u").first

  # A default set of traceroute and ping options, in the same format as the
  # {options} argument of execute_trace and execute_ping.  Any explicitly
  # provided options will override default options with the same key.
  attr_accessor :default_trace_options, :default_ping_options

  attr_accessor :debug_data, :debug_commands

  # The {format} argument should be "warts" or "json", and it's only
  # supported by CAIDA's custom scamper build for the moment (2017-06-09).
  # This specifies the format of the data returned by scamper over the
  # control socket.
  #
  # The {syslog} argument should be an instance of Syslog.
  def initialize(scamperio, format=nil, syslog=nil)
    @scamperio = scamperio
    @scamperio.client = self
    @log = syslog || StderrPrinter.new
    @warts_format = (!format || format == "warts")

    @requested_shutdown = false
    @awaiting_shutdown = false  # true | false | "sent"
    @unacked_commands = []  # [command, target, options]
    @unsent_commands = []   # [command, target, options]
    @scamper_wants_more = false
  end


  # This callback is triggered whenever scamper returns an OK response.
  # This is not triggered for the shutdown command.
  #
  # The block should have the signature |target, options|, where
  # 'target' and 'options' correspond to arguments originally supplied
  # to execute_trace, execute_ping, or execute_raw_command ('options' will
  # be nil for the latter).
  def on_success(&block)
    @on_success_handler = block
  end


  # This callback is triggered whenever scamper returns an ERROR response.
  # This is not triggered for the shutdown command.
  #
  # The block should have the signature |message, target, options|,
  # where 'message' is the error message returned by scamper, and
  # 'target' and 'options' correspond to arguments originally supplied
  # to execute_trace, execute_ping, or execute_raw_command ('options' will
  # be nil for the latter).
  def on_error(&block)
    @on_error_handler = block
  end


  # This callback is triggered if, and only if, scamper is ready for more
  # commands and no unsent commands are available.  The user, however, is
  # not required to submit any new commands in this callback; this is
  # merely a signal letting the user know that new ones can be submitted
  # if any are available.
  #
  # The user may accumulate unsent commands by calling
  # execute_trace/execute_ping/execute_raw_command from this callback
  # or from outside this callback (for example, some big job
  # submission at program start up).
  #
  # No arguments are passed to the block.
  def on_more(&block)
    @on_more_handler = block
  end


  # This callback is triggered whenever scamper returns warts data over the
  # control socket.
  #
  # The block receives exactly one warts element (in its entirety), and the
  # client should be prepared to handle all types of records (addresses,
  # traces, pings, dealiases, lists, and cycles).
  #
  # The block can write out the received element $e to a warts file $f
  # opened with Warts::File.open with
  #
  #    $f.write $e
  #
  # This is the preferred way of saving warts data on the client side.
  # However, if you wish, you can also save out warts data from the
  # on_raw_data handler.
  #
  # The element will be an instance of Warts::Trace, Warts::List, etc.
  # of rb-wartslib.
  def on_data(&block)
    @on_data_handler = block
  end


  # This callback is triggered whenever scamper returns warts data over the
  # control socket.
  #
  # The block should have the signature |binary_record, uuencoded_data|,
  # where
  #
  # 'binary_record' is the raw binary warts record which you can
  #     directly append to a file (a simple binary output file and not
  #     one opened with Warts::File.open--see on_data) to create a
  #     well-formed warts file (however, if you take this route, then
  #     you must handle and write out all such binary records or
  #     you'll get a malformed warts file), and
  #
  # 'uuencoded_data' is the uuencoded form of the above, in the format
  #     returned by scamper over the control socket (that is, "DATA
  #     \d+\n<uuencoded text lines>\n").
  #
  # The warts record (and its uuencoded form) is provided in its entirety.
  # The client should be prepared to handle all types of records (addresses,
  # traces, pings, dealiases, lists, and cycles).
  def on_raw_data(&block)
    @on_raw_data_handler = block
  end


  # This callback is triggered when the connection to scamper is lost (but
  # not at shutdown) or some error occurs in ScamperIO or ScamperDataFeed.
  #
  # The block receives an error message.
  def on_service_failure(&block)
    @on_service_failure_handler = block
  end


  # This callback is triggered when the session with scamper has been
  # cleanly shutdown after a shutdown request from the client.
  #
  # No arguments are passed to the block.
  def on_shutdown(&block)
    @on_shutdown_handler = block
  end


  #==========================================================================

  # Schedules a traceroute command.
  #
  # ScamperIO will take care of actually issuing the command to scamper at
  # the right time--that is, when scamper is ready for the next command.
  # Hence, the client may call this method at any time, even outside the
  # on_more() callback.
  #
  # Returns the actual "trace" command issued to scamper.
  # Raises ArgumentError if options are malformed.
  def execute_trace(target, options={})
    # propagate out an ArgumentError or other exception on malformed options
    trace_command = generate_trace_command target, options
    submit_scamper_command trace_command, target, options
    trace_command
  end


  # Schedules a ping command.
  #
  # ScamperIO will take care of actually issuing the command to scamper at
  # the right time--that is, when scamper is ready for the next command.
  # Hence, the client may call this method at any time, even outside the
  # on_more() callback.
  #
  # Returns the actual "ping" command issued to scamper.
  # Raises ArgumentError if options are malformed.
  def execute_ping(target, options={})
    # propagate out an ArgumentError or other exception on malformed options
    ping_command = generate_ping_command target, options
    submit_scamper_command ping_command, target, options
    ping_command
  end


  # Schedules a raw scamper command.
  #
  # The 'target' parameter is any arbitrary value useful to the client for
  # identifying the raw command.  The 'target' will be returned in various
  # callbacks.
  #
  # ScamperIO will take care of actually issuing the command to scamper at
  # the right time--that is, when scamper is ready for the next command.
  # Hence, the client may call this method at any time, even outside the
  # on_more() callback.
  def execute_raw_command(command, target)
    submit_scamper_command command, target
    nil
  end


  # Schedules shutdown of the scamper session, which will end once all
  # scheduled commands have been issued and their results retrieved.
  #
  # The client may call this method at any time, even outside the
  # on_more() callback.
  def shutdown
    return if @requested_shutdown
    @requested_shutdown = true
    unless @awaiting_shutdown
      @awaiting_shutdown = true
      if @unsent_commands.empty?
        @scamperio.write_command "done"
        @awaiting_shutdown = "sent"
      end

      # else if @unsent_commands is not empty,
      # the 'done' command will be issued in process_scamper_more_response()
      # because @awaiting_shutdown == true (rather than == 'sent')
    end
  end


  #==========================================================================

  # ScamperIO event handler
  def process_scamper_response(response)
    if @debug_data
      @log.debug ">> unacked_commands=%d, unsent_commands=%d, " \
        "scamper_wants_more=%p", @unacked_commands.size, @unsent_commands.size,
        @scamper_wants_more
    end

    case response
    # Note: Older scamper versions return "OK\n"; newer "OK id-1\n".
    when /^OK/ then process_scamper_ok_response()

    when /^ERR /
      process_scamper_error_response response

      # scamper doesn't return a 'MORE' response after the 'ERR' message for
      # an invalid traceroute, even if scamper was in the 'MORE' state prior
      # to the issuance of the bad command, so we simulate it.
      process_scamper_more_response()

    when "MORE\n" then process_scamper_more_response()
    when /^DATA \d/ then process_scamper_data_response response
    else
      msg = sprintf "ScamperDataFeed: got unknown scamper response: %p",
        response
      @log.err "%s", msg
      @on_service_failure_handler.call msg if @on_service_failure_handler
    end
  end


  # ScamperIO event handler
  def lost_scamper_connection()
    if @awaiting_shutdown == "sent"
      @on_shutdown_handler.call() if @on_shutdown_handler
    else
      if @on_service_failure_handler
        @on_service_failure_handler.call "lost connection to scamper"
      end        
    end
  end


  # ScamperIO event handler
  def scamperio_failed(message)
    @on_service_failure_handler.call message if @on_service_failure_handler
  end


  #==========================================================================
  private

  def process_scamper_ok_response
    unless @awaiting_shutdown == "sent"
      entry = @unacked_commands.shift
      if entry && @on_success_handler
        command, target, options = entry
        @on_success_handler.call target, options
      end
    end
  end


  def process_scamper_error_response(response)
    if @awaiting_shutdown == "sent"
      # XXX
    else
      error_message = /^ERR (.*)$/.match(response)[1].chomp
      entry = @unacked_commands.shift
      if entry && @on_error_handler
        command, target, options = entry
        @on_error_handler.call error_message, target, options
      end
    end
  end


  def process_scamper_more_response
    call_handler = false
    if @unsent_commands.empty?
      if @awaiting_shutdown
        # if @awaiting_shutdown == true, then the client has only requested it
        unless @awaiting_shutdown == "sent"
          @awaiting_shutdown = "sent"
          @scamperio.write_command "done"
        end
      else
        @scamper_wants_more = true
        call_handler = true
      end
    else
      entry = @unsent_commands.shift
      @unacked_commands << entry
      @scamperio.write_command entry[0]
    end

    if call_handler && @on_more_handler
      @on_more_handler.call()
    end
  end


  def process_scamper_data_response(response)
    @log.debug "ScamperDataFeed: %p", response if @debug_data
    data = []
    response.each_line do |line|
      next if line =~ /^DATA \d+$/
      if @warts_format
        data << line.unpack("u")
      else
        data << line  # JSON output isn't uuencoded
      end
    end

    bin_data = data.join("")

    if @warts_format
      process_warts_objects bin_data, response
    else
      @on_raw_data_handler.call(bin_data, response) if @on_raw_data_handler
      @on_data_handler.call bin_data if @on_data_handler
    end
  end


 def process_warts_objects(bin_data, response)
    header = bin_data.slice 0, 4
    # skip over List and Cycle objects
    # XXX This assumes each bin_data only has a single warts object!  True??
    return if header == "\x12\x05\x00\x01" || header == "\x12\x05\x00\x02"

    warts_tmpfile = Tempfile.new "scamperdatafeed-#{$$}-"
    warts_tmpfile.write WARTS_LIST_OBJECT
    warts_tmpfile.write WARTS_CYCLE_OBJECT
    warts_tmpfile.write bin_data
    warts_tmpfile.close

    Warts::File.open(warts_tmpfile.path) do |file|
      file.add_filters Warts::TRACE, Warts::PING
      file.read do |element|
        @log.debug "ScamperDataFeed: %p", element if @debug_data
        @on_raw_data_handler.call(bin_data, response) if @on_raw_data_handler
        @on_data_handler.call element if @on_data_handler
      end
    end

    warts_tmpfile.unlink()
  end


  def submit_scamper_command(command, target, options)
    # Stop scheduling commands if shutdown has been requested/executed.
    return if @awaiting_shutdown  

    # Push in @unsent_commands first so that commands are always executed
    # in order.  Usually, if @scamper_wants_more is true, then there won't
    # be any unsent commands, but it's better to be safe than sorry, given
    # the complications caused by executing the on_more handler upon
    # an ERR response (see handle_process_scamper_response).
    @unsent_commands << [command, target, options]
    if @scamper_wants_more
      @scamper_wants_more = false
      entry = @unsent_commands.shift
      @unacked_commands << entry
      @scamperio.write_command entry[0]
    end
  end


  SCAMPER_TRACE_OPTIONS = {
    "pmtu" => [ "-M", :none ],
    "allattempts" => [ "-Q", :none ],
    "continueprobing" => [ "-T", :none ],

    "confidence" => [ "-c", :numeric ],
    "dport" => [ "-d", :numeric ],
    "firsthop" => [ "-f", :numeric ],
    "gaplimit" => [ "-g", :numeric ],
    "gapaction" => [ "-G", :numeric ],
    "loops" => [ "-l", :numeric ],
    "loopaction" => [ "-L", :numeric ],
    "maxttl" => [ "-m", :numeric ],
    "payload" => [ "-p", :string ],
    "method" => [ "-P", :string ],
    "attempts" => [ "-q", :numeric ],
    "sport" => [ "-s", :numeric ],
    "tos" => [ "-t", :numeric ],
    "userid" => [ "-U", :numeric ],
    "wait" => [ "-w", :numeric ],
    "waitprobe" => [ "-W", :numeric ]
  }

  SCAMPER_PING_OPTIONS = {
    "record_route" => [ "-R", :none ],

    "payload" => [ "-B", :string ],
    "attempts" => [ "-c", :numeric ],
    "icmpsum" => [ "-C", :numeric ],
    "dport" => [ "-d", :numeric ],
    "sport" => [ "-F", :numeric ],
    "waitprobe" => [ "-i", :numeric ],
    "ttl" => [ "-m", :numeric ],
    "required_replies" => [ "-o", :numeric ],
    "pattern" => [ "-p", :string ],
    "method" => [ "-P", :string ],
    "size" => [ "-s", :numeric ],
    "timestamp" => [ "-T", :string ],
    "userid" => [ "-U", :numeric ],
    "timeout" => [ "-W", :numeric ],
    "tos" => [ "-z", :numeric ]
  }


  def generate_trace_command(target, options)
    generate_trace_or_ping_command "trace", target, options,
      @default_trace_options, SCAMPER_TRACE_OPTIONS
  end


  def generate_ping_command(target, options)
    generate_trace_or_ping_command "ping", target, options,
      @default_ping_options, SCAMPER_PING_OPTIONS
  end


  # The keys in {options} and {default_options} can be either symbols or
  # strings.
  #
  # The "user_data" key is ignored and may be used by the client to
  # associate state with this target that can be retrieved in on_success
  # or on_error.
  def generate_trace_or_ping_command(command_name, target, options,
                                     default_options, option_mapping)
    command = [ command_name ]
    begin
      if default_options
        options = default_options.merge options
      end

      options.each do |key, value|
        if key == "raw_options"
          command << value
        elsif key == "user_data"
          # nothing to do
        else
          option_name, option_type = option_mapping[key.to_s]
          unless option_name
            raise ArgumentError,
              "unknown #{command_name} option '" + key.to_s + "'"
          end

          command << option_name
          case option_type
          when :none
            # nothing to do
          when :numeric
            command << value.to_s
          when :string
            s = value.to_s
            if s.index("\n") || s.index("\r")
              raise ArgumentError,
                "embedded newline or carriage return in value of option '" +
                key.to_s + "'"
            end
            command << s.inspect
          end
        end
      end
    end
    command << target
    command.join " "
  end

end

end  # module ArkUtil
