#############################################################################
## This class provides SFTP operations that time out when a network transfer
## stalls or slows down below some specified rate.
##
## --------------------------------------------------------------------------
## Copyright (C) 2007-2008 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: timedsftp.rb,v 1.40 2010/02/23 19:47:25 youngh Exp $
#############################################################################

require 'net/sftp'
require 'logger'

module ArkUtil

class TimedSFTP

  attr_accessor :debug_sftp

  class StderrPrinter
    def method_missing(name, *args)
      args[0] = name.to_s.upcase! + ": " + args[0] + "\n"
      $stderr.printf *args
    end
  end

  # Data transfer is subject to two levels of timeouts and retries, an outer
  # 'slow' retry and an inner 'fast' retry, as follows:
  #
  #   -------------------------------------------------------------------
  #   * attempt 1
  #
  #     + attempt A
  #     + 'fast' retry: attempt B, retrying after 4 seconds
  #     + 'fast' retry: attempt C, retrying after 16 seconds
  #     + 'fast' retry: attempt D, retrying after 64 seconds
  #
  #   * 'slow' retry: attempt 2, retrying after 15 minutes
  #
  #     + attempt A
  #     + 'fast' retry: attempt B, retrying after 4 seconds
  #     + 'fast' retry: attempt C, retrying after 16 seconds
  #     + 'fast' retry: attempt D, retrying after 64 seconds
  #
  #   * 'slow' retry: attempt 3, retrying after 60 minutes
  #      ...
  #
  #   * 'slow' retry: attempt 4, retrying after 4 hours
  #      ...
  #
  #   * 'slow' retry: attempt 5, 6, ..., N, retrying after 16 hours
  #      ...
  #   -------------------------------------------------------------------
  #
  # The 'slow' retry attempts are spread relatively far apart and are meant
  # to be on the timescale of equipment failure-fix cycles.  The 'fast'
  # retry attempts are meant to be on the timescale of transient network
  # problems (e.g., temporary routing loops, congestion-based loss/collapse,
  # route loss during BGP path exploration).
  #
  # Retries happen for two reasons, failure to open a connection and failure
  # to make progress on a transfer.  Currently, "failure to make progress" is
  # defined simplisitically as any period of N consecutive seconds with 0
  # bytes transferred.  A more sophisticated definition might be based on
  # a lower bound on the transfer rate as computed with an exponential moving
  # average.
  #
  # NOTE: If a transfer fails to make progress, the retry attempt happens
  #       in the outer 'slow' retry loop rather than in the inner 'fast'
  #       retry loop.  The purpose of the inner 'fast' retry loop is to
  #       react quickly to connection loss and failure to open a connection.
  #       If a transfer fails to make progress, then we typically want to
  #       react more slowly (that is, retry after a moderately long delay),
  #       so the inner 'fast' retry loop is aborted, and the transfer is
  #       retried in the outer 'slow' loop.  (An important point is that
  #       failure to make progress is detected at all times, regardless of
  #       at what stage the transfer is at.)
  # 
  # Failure to open a connection ("open-conn failure") is handled separately
  # from the other failures that happen once a connection is opened and the
  # transfer started.  An open-conn failure is retried at a different
  # configured retry rate than either the 'slow' or 'fast' retry rates.
  # Furthermore, an open-conn failure is retried indefinitely, whereas the
  # 'slow' and 'fast' retries are limited to a finite number of attempts.
  # Retries for open-conn failures aren't counted toward the 'slow' retries
  # even though retries for open-conn failures happen at the same level as
  # 'slow' retries, as shown below:
  #
  #   -------------------------------------------------------------------
  #   * attempt 1
  #
  #     + attempt A
  #     + 'fast' retry: attempt B, retrying after 4 seconds
  #     + 'fast' retry: attempt C, retrying after 16 seconds
  #     + 'fast' retry: attempt D, retrying after 64 seconds
  #
  #      <<<connection couldn't be opened even after 3 'fast' retries>>>
  #
  #   * retrying open connection after 15 minutes
  #
  #     + attempt A
  #     + 'fast' retry: attempt B, retrying after 4 seconds
  #     + 'fast' retry: attempt C, retrying after 16 seconds
  #     + 'fast' retry: attempt D, retrying after 64 seconds
  #
  #      <<<connection couldn't be opened even after 3 'fast' retries>>>
  #
  #   * retrying open connection after 60 minutes
  #
  #     + attempt A
  #     + 'fast' retry: attempt B, retrying after 4 seconds
  #     + 'fast' retry: attempt C, retrying after 16 seconds
  #     + 'fast' retry: attempt D, retrying after 64 seconds
  #
  #      <<<connection couldn't be opened even after 3 'fast' retries>>>
  #
  #   * retrying open connection after 4 hours
  #      ...
  #
  #      <<<some failure besides failure to open connection>>>
  #
  #   * 'slow' retry: attempt 2, retrying after 15 minutes
  #
  #     + attempt A
  #     + 'fast' retry: attempt B, retrying after 4 seconds
  #     + 'fast' retry: attempt C, retrying after 16 seconds
  #     + 'fast' retry: attempt D, retrying after 64 seconds
  #
  #   -------------------------------------------------------------------
  #
  # NOTE: 'fast' and 'slow' retries are based on different

  #=========================================================================

  # Parameters for 'slow' retries.
  #
  # The idle timeout starts at INITIAL_IDLE_TIMEOUT and increases by a
  # factor of IDLE_TIMEOUT_MULTIPLIER at each attempt, capping at
  # MAX_IDLE_TIMEOUT.  No more than MAX_SLOW_ATTEMPTS attempts will
  # be made.
  #
  # Although not strictly required, it makes the most sense for the
  # parameters for 'slow' retries to be larger than those for 'fast' retries
  # (given below).  In particular, INITIAL_IDLE_TIMEOUT should be greater
  # than the total amount of delay possible with 'fast' retries after the
  # maximum allowed number of retries.  If INITIAL_IDLE_TIMEOUT is set too
  # low, then it is possible for the 'slow' retry loop to interrupt the
  # 'fast' retry loop.  This could be the desired behavior in some cases,
  # but in most cases, it probably will not be.
  #
  # The WATCHDOG_TIMEOUT specifies the interval at which to monitor download
  # progress, namely by checking the current size of the local file to which
  # downloaded content is being written.  This timeout should be significantly
  # lower than INITIAL_IDLE_TIMEOUT but not too low as to create an undue load
  # on the system.  Furthermore, there is a point of diminishing return in the
  # responsiveness provided by a low WATCHDOG_TIMEOUT.  It is recommended
  # that WATCHDOG_TIMEOUT be no less than 60 seconds, which strikes a nice
  # balance.

  PRODUCTION_PARAMETERS = true

  if PRODUCTION_PARAMETERS
    INITIAL_IDLE_TIMEOUT = 15 * 60 # seconds
    MAX_IDLE_TIMEOUT = 2 * 3600 # seconds
    IDLE_TIMEOUT_MULTIPLIER = 2
    MAX_SLOW_ATTEMPTS = 15  # around 24 hours max attempt duration
    WATCHDOG_TIMEOUT = 60 # seconds
  else
    # values for testing purposes
    INITIAL_IDLE_TIMEOUT = 60 # seconds
    MAX_IDLE_TIMEOUT = 120 # seconds
    IDLE_TIMEOUT_MULTIPLIER = 2
    MAX_SLOW_ATTEMPTS = 3
    WATCHDOG_TIMEOUT = 10 # seconds
  end

  # Parameters for handling failures to open connections (open-conn failures).
  #
  # Unlike 'slow' and 'fast' retries, open-conn failures are retried
  # indefinitely.  The delay between retries starts at INITIAL_CONNECT_DELAY
  # and increases by a factor of CONNECT_DELAY_MULTIPLIER, capping at
  # MAX_CONNECT_DELAY.

  if PRODUCTION_PARAMETERS
    INITIAL_CONNECT_DELAY = 120 # seconds
    MAX_CONNECT_DELAY = 240 # seconds
    CONNECT_DELAY_MULTIPLIER = 2
  else
    # values for testing purposes
    INITIAL_CONNECT_DELAY = 15 # seconds
    MAX_CONNECT_DELAY = 30 # seconds
    CONNECT_DELAY_MULTIPLIER = 2
  end

  # Parameters for 'fast' retries.
  #
  # The delay between retries starts at INITIAL_FAST_RETRY_DELAY and
  # increases by a factor of FAST_RETRY_DELAY_MULTIPLIER at each attempt,
  # capping at MAX_FAST_RETRY_DELAY.  No more than MAX_FAST_ATTEMPTS
  # attempts will be made.
  #
  # Generally speaking, these parameter values should be lower than the
  # values for 'fast' retries.  See the comments at the latter for details.

  INITIAL_FAST_RETRY_DELAY = 4 # seconds
  MAX_FAST_RETRY_DELAY = 16 # seconds
  FAST_RETRY_DELAY_MULTIPLIER = 4
  MAX_FAST_ATTEMPTS = 3


  #=========================================================================

  # Exit codes:
  #
  #   0 -- everything was successful; the only possible non-success is a
  #        warning that the file couldn't be removed from the server
  #
  #   1 -- file doesn't exist on the server
  #
  #   2 -- file downloaded successfully but some error happened afterwards
  #        (e.g., couldn't rename <file>.download to <file>); don't retry
  #        download, and generally prefer to fail hard and defer to human
  #        intervention
  #
  #   3 -- file not downloaded successfully and some non-fatal error happened;
  #        may retry
  #
  #   4 -- file not downloaded successfully and some fatal error happened;
  #        don't retry
  EXIT_SUCCESS = 0
  EXIT_NO_FILE = 1
  EXIT_PARTIAL_SUCCESS_DONT_RETRY = 2
  EXIT_FAILED_BUT_RETRY = 3
  EXIT_FAILED_DONT_RETRY = 4
  EXIT_COULDNT_CONNECT_BUT_RETRY = 5

  #=========================================================================

  # {ssh_options} can contain any of the options supported by Net::SFTP.start,
  # such as :user, :password, and :port.
  #
  #   NOTE: Net::SSH v1.x used :username and not :user.  For backwards
  #         compatibility, this class accepts either.
  #
  # {options} are options to TimedSFTP; supported options:
  #
  #   :delete_downloaded => true [default: false]
  #      whether to remove a file from the server after successful download;
  #
  def initialize(host, ssh_options={}, syslog=nil, options={})
    @host = host
    @ssh_options = ssh_options
    @full_host = ssh_full_host host, ssh_options
    @log = syslog || StderrPrinter.new
    @options = options  # hash of symbols to values

    # Smooth over the transition from Net::SSH v1 to v2.
    if @ssh_options.has_key? :username
      @ssh_options[:user] = @ssh_options[:username]
      @ssh_options.delete :username
    end

    unless @ssh_options.has_key? :user
      raise ArgumentError, "missing :user SSH option"
    end
  end


  # Returns * true on success,
  #         * false if the remote file doesn't exist, and
  #         * {error string} if transfer fatally failed, possibly after
  #           multiple retries.
  #
  # This method prefers to exit on certain errors rather than trying to
  # recover, since it's best for a human to intervene in some cases, even
  # when the error wouldn't necessarily be considered fatal by everyone.
  # This "fail-fast" approach does, however, depend on having multiple
  # collection processes running concurrently, so that data collection
  # doesn't completely halt.

  def get_file(remote_file, local_file)
    local_file_temp = "#{local_file}.download.#{$$}"
    connect_delay = INITIAL_CONNECT_DELAY
    timeout = INITIAL_IDLE_TIMEOUT
    current_try = 1
    loop do
      catch(:attempt) do
	begin
	  File.unlink local_file_temp
	rescue SystemCallError
	  unless Errno::ENOENT === $!
	    msg = ("TimedSFTP#get_file(%s:%s): couldn't unlink '%s': %s: %s" %
		   [@full_host, remote_file, local_file_temp, $!.class.name,
		     $!.to_s])
	    @log.err("%s", msg)
	    return msg
	  end
	end

	pid = fork do
          begin
            get_file_process remote_file, local_file, local_file_temp
          rescue
            @log.err("TimedSFTP#get_file_process(%s:%s): unexpected " +
                     "exception: %s: %s; backtrace: %s", @full_host,
                     remote_file, $!.class.name, $!.to_s,
                     $!.backtrace.join(" <= "))
            exit! EXIT_FAILED_DONT_RETRY
          end
	end
	@log.debug("TimedSFTP#get_file(%s:%s): get_file_process pid=%d",
		   @full_host, remote_file, pid)

	attempt_time = Time.now
	last_size = File.size(local_file_temp) rescue 0
	last_changed = Time.now

        # Note: An ensure clause uses the value of the {child_status} variable
        #       to determine whether it needs to clean up the get_file child
        #       process.  If the {child_status} variable is nil, then the
        #       child process needs to be terminated and waited for.
	child_status = nil
	begin
          watchdog_time = Time.now
          loop do
            pid_result, child_status = Process.waitpid2 pid, Process::WNOHANG
            break if pid_result

            if Time.now - watchdog_time < WATCHDOG_TIMEOUT
              sleep_at_least 1
            else
              size = File.size(local_file_temp) rescue 0
              if size != last_size
                last_size = size
                last_changed = Time.now
              end

              @log.debug("TimedSFTP#get_file(%s:%s): in progress: size=%d, " +
                         "last_changed=%d (%s)", @full_host, remote_file,
                         size, last_changed, last_changed)

              unless Time.now - last_changed > timeout
                sleep_at_least 1
                watchdog_time = Time.now
                next  # retry waitpid
              end

              if current_try == MAX_SLOW_ATTEMPTS
                msg = ("TimedSFTP#get_file(%s:%s): reached maximum number " +
                       "of 'slow' attempts; aborting") %
                       [@full_host, remote_file]
                @log.err("%s", msg)
                return msg
              else
                current_try += 1
                timeout = [MAX_IDLE_TIMEOUT,
                           timeout * IDLE_TIMEOUT_MULTIPLIER].min
                @log.info("TimedSFTP#get_file(%s:%s): doing 'slow' retry " +
                          "attempt %d immediately", @full_host, remote_file,
                          current_try)
                throw :attempt
              end
            end
          end

	  if child_status.exited?
	    case child_status.exitstatus
	    when EXIT_SUCCESS
	      return true

	    when EXIT_NO_FILE
	      File.unlink local_file_temp rescue nil
	      return false

	    when EXIT_PARTIAL_SUCCESS_DONT_RETRY
	      msg = ("TimedSFTP#get_file(%s:%s): downloaded successfully" +
		     " to '%s' but couldn't rename to final name; " +
		     " manually rename file; aborting") %
		     [@full_host, remote_file, local_file_temp]
	      @log.err("%s", msg)
	      return msg

	    when EXIT_FAILED_BUT_RETRY
	      @log.info("TimedSFTP#get_file(%s:%s): get_file_process failed " +
			"but we can retry", @full_host, remote_file)

	      if current_try == MAX_SLOW_ATTEMPTS
		msg = ("TimedSFTP#get_file(%s:%s): reached maximum " +
		       "number of 'slow' attempts; aborting") %
		       [@full_host, remote_file]
		@log.err("%s", msg)
		return msg
	      else
		current_try += 1
		delay = [0, timeout - (Time.now - attempt_time)].max
		@log.info("TimedSFTP#get_file(%s:%s): doing 'slow' retry " +
			  "attempt %d after %d seconds", @full_host,
			  remote_file, current_try, delay)

		sleep_at_least delay
		timeout = [MAX_IDLE_TIMEOUT,
		           timeout * IDLE_TIMEOUT_MULTIPLIER].min
		throw :attempt
	      end

	    when EXIT_FAILED_DONT_RETRY
	      msg = ("TimedSFTP#get_file(%s:%s): get_file_process had " +
		     "a fatal error; aborting") % [@full_host, remote_file]
	      @log.err("%s", msg)
	      return msg

	    when EXIT_COULDNT_CONNECT_BUT_RETRY
	      @log.info("TimedSFTP#get_file: couldn't connect to '%s'; " +
			"retrying after %d seconds", @full_host, connect_delay)

	      sleep_at_least connect_delay
	      connect_delay = [MAX_CONNECT_DELAY, 
		               connect_delay * CONNECT_DELAY_MULTIPLIER].min
	      throw :attempt
	    end
	  end

	  msg = ("TimedSFTP#get_file(%s:%s): get_file_process exited " +
		 "with an unknown status: %p; aborting") %
	         [@full_host, remote_file, child_status]
	  @log.err("%s", msg)
	  return msg

	ensure
	  unless child_status
	    @log.debug("TimedSFTP#get_file(%s:%s): killing get_file_process" +
		       " (pid=%d)", @full_host, remote_file, pid)
            Process.kill "KILL", pid  # killing an already dead process is ok
            Process.waitpid pid
	    @log.debug("TimedSFTP#get_file(%s:%s): get_file_process (pid=%d)" +
		       " exited with %p", @full_host, remote_file, pid, $?)
	  end
	end
      end
    end
  end


  private #-----------------------------------------------------------------

  # NOTE: Processes must exit with exit!() rather than exit(), since the
  #       latter may cause resources duplicated from the parent process
  #       to be cleaned up, which might be problematic and unexpected from
  #       the user's point of view.

  def get_file_process(remote_file, local_file, local_file_temp)
    delay = INITIAL_FAST_RETRY_DELAY # seconds
    current_try = 1
    begin
      # NOTE: The username must be provided via @ssh_options[:user].
      Net::SFTP.start(@host, @ssh_options[:user], @ssh_options) do |sftp|
	begin
	  download_file sftp, remote_file, local_file_temp
	rescue EOFError, SystemCallError
	  @log.err("TimedSFTP#get_file_process(%s:%s): error downloading: " +
		   "%s: %s", @full_host, remote_file, $!.class.name, $!.to_s)
          if SystemCallError === $!
            # XXX Not sure what the full set of possible errors is here.
            #     It's at least Errno::EPIPE and Errno::ECONNRESET.
            #     We're taking a slight risk in retrying without knowing
            #     the full set of possible errors.
            exit! EXIT_FAILED_BUT_RETRY
          else
            exit! EXIT_FAILED_DONT_RETRY
          end

	rescue Net::SFTP::StatusException
	  if $!.code == Net::SFTP::Constants::StatusCodes::FX_NO_SUCH_FILE
	    @log.info("TimedSFTP#get_file_process: remote file '%s:%s' " +
		      "doesn't exist", @full_host, remote_file)
	    exit! EXIT_NO_FILE
	  else
	    raise
	  end
	end

	begin
	  File.rename local_file_temp, local_file
	rescue SystemCallError
	  @log.err("TimedSFTP#get_file_process: couldn't rename '%s' to " +
		   "'%s': %s",	   local_file_temp, local_file, $!.to_s)
	  exit! EXIT_PARTIAL_SUCCESS_DONT_RETRY
	end

	begin
	  sftp.remove! remote_file if @options[:delete_downloaded]
	rescue Net::SFTP::Exception, Errno::EPIPE
	  @log.warning("TimedSFTP#get_file_process: couldn't remove '%s:%s' " +
		       "on remote server; manually remove to avoid possible " +
		       "accidental re-downloading: %s",
		       @full_host, remote_file, $!.to_s)
	end

	exit! EXIT_SUCCESS
      end

    rescue Net::SSH::AuthenticationFailed
      @log.err("TimedSFTP#get_file_process: couldn't connect to '%s': %s: %s",
	       @full_host, $!.class.name, $!.to_s)
      exit! EXIT_FAILED_DONT_RETRY

    # EOFError:
    # (backtrace in .../net-ssh-2.0.8/lib/net/ssh/transport/):
    #   .../server_version.rb:44:in `readline'
    #   .../server_version.rb:44:in `negotiate!'
    #   .../server_version.rb:43:in `loop'
    #   .../server_version.rb:43:in `negotiate!'
    #   .../server_version.rb:32:in `initialize'
    #   .../session.rb:75:in `new'

    # "IOError: closed stream":
    # (backtrace in .../net-ssh-2.0.8/lib/net/ssh/transport/):
    #   .../packet_stream.rb:68:in `select'
    #   .../packet_stream.rb:68:in `available_for_read?'
    #   .../packet_stream.rb:80:in `next_packet'
    #   .../session.rb:160:in `poll_message'
    #   .../session.rb:155:in `loop'
    #   .../session.rb:155:in `poll_message'
    #
    # also in .../net/ssh/buffered_io.rb:98:in `send'

    # "Net::SSH::Disconnect: connection closed by remote host":
    # often (backtrace in .../net-ssh-2.0.8/lib/net/ssh/transport/):
    #   .../packet_stream.rb:94:in `next_packet'
    #   .../packet_stream.rb:84:in `loop'
    #   .../packet_stream.rb:84:in `next_packet'
    #   .../session.rb:160:in `poll_message'
    #   .../session.rb:155:in `loop'
    #   .../session.rb:155:in `poll

    # XXX: [[Pre-Net::SFTP v2.0.0 comments but might be still relevant.]]
    #      For some weird reason, I can't 'rescue' on
    #      Net::SFTP::Operations::StatusException--I get NameError.
    #      StatusException is probably dynamically created somehow.
    rescue SocketError, SystemCallError, EOFError, IOError,
      Net::SFTP::Exception, Net::SSH::Exception
      case $!
      # NOTE: SocketError is thrown when the hostname is invalid (or rather
      #       couldn't be looked up).  Hence, there's a chance for a
      #       non-recoverable problem, like a typo in the hostname.  The
      #       problem could also be something transient like a temporary
      #       DNS failure, and we retry for that reason.
      #
      # ECONNABORTED (Software caused connection abort) --
      #        failed to establish connection: got RST instead of SYN/ACK
      when SocketError, Errno::ECONNABORTED, Errno::ECONNREFUSED,
	  Errno::EHOSTDOWN, Errno::EHOSTUNREACH, Errno::ENETDOWN,
	  Errno::ENETUNREACH
	@log.err("TimedSFTP#get_file_process: couldn't connect to '%s': " +
		 "%s: %s", @full_host, $!.class.name, $!.to_s)
	exit_code = EXIT_COULDNT_CONNECT_BUT_RETRY

      # ENETRESET (Network dropped connection [on | because of] reset) --
      #        local network dropped connection when interface was reset (?)
      when Errno::ECONNRESET, Errno::ENETRESET, EOFError, IOError
	@log.err("TimedSFTP#get_file_process: lost connection to '%s': %s: %s",
		 @host, $!.class.name, $!.to_s)
	exit_code = EXIT_FAILED_BUT_RETRY

      else # when Net::SFTP::Exception and Net::SSH::Exception:
        # Net::SFTP.start raises 'Errno::EINVAL: Invalid argument - connect(2)'
        # when it is unable to connect to the destination due to a transient
        # failure of some kind (it's not DNS, since we always connect with
        # IP addresses).  This exception is reported here.
	@log.err("TimedSFTP#get_file_process(%s:%s): got exception %s: %s; " +
                 "backtrace: %s", @full_host, remote_file, $!.class.name,
                 $!.to_s, $!.backtrace.join(" <= "))
	exit_code = EXIT_FAILED_BUT_RETRY
      end

      if current_try == MAX_FAST_ATTEMPTS
	@log.info("TimedSFTP#get_file_process(%s:%s): reached maximum " +
		  "number of 'fast' attempts; aborting",
		  @full_host, remote_file)
	exit! exit_code
      else
	current_try += 1
	@log.info("TimedSFTP#get_file_process(%s:%s): doing 'fast' retry " +
		  "attempt %d after %d seconds", @full_host, remote_file,
		  current_try, delay)

	sleep_at_least delay
	delay = [MAX_FAST_RETRY_DELAY, delay * FAST_RETRY_DELAY_MULTIPLIER].min
	retry
      end
    end
  end


  # Downloads a remote file to a local file.
  #
  # This returns normally if everything is successful.  Otherwise, an
  # exception will be thrown, and it is up to the caller to handle these
  # exceptions.  The most important exceptions are
  #
  #   Errno::EPIPE: Broken pipe - send(2) => lost SFTP connection (?)
  #
  #   SystemCallError => local I/O error (?)
  #
  #   Net::SFTP::StatusException
  #      with code == Net::SFTP::Constants::StatusCodes::FX_NO_SUCH_FILE (2)
  #                   => the remote file doesn't exist
  # 
  def download_file(sftp, remote_file, local_file, options={})
    # Note: Net::SFTP v2.0.0 doesn't support a :preserve option for #download,
    #       so we have to manually set the file attributes.
    #
    # Since sftp.download! raises an inconvenient RuntimeError on missing
    # remote file, we call sftp.stat! to trigger Net::SFTP::StatusException
    # instead.
    attributes = sftp.stat! remote_file
    if @debug_sftp
      puts "downloading"
      logger = Logger.new STDOUT
      logger.level = Logger::DEBUG
      sftp.logger = logger

      # NOTE: Use read_size of 16000 as workaround for some bug, either in
      #       net-sftp or in remote sftp-server.  The default read size of
      #       32000 is problematic; it causes transfers to hang for certain
      #       remote file sizes.
      sftp.download!(remote_file, local_file, :read_size => 16000) do
        |event, downloader, *args|
        case event
        when :open then
          # args[0] : file metadata
          puts "starting download: #{args[0].remote} -> #{args[0].local} " +
            "(#{args[0].size} bytes}"
          p args[0]
        when :get then
          # args[0] : file metadata
          # args[1] : byte offset in remote file
          # args[2] : data that was received
          puts "writing #{args[2].length} bytes to #{args[0].local} " +
            "starting at #{args[1]}"
        when :close then
          # args[0] : file metadata
          puts "finished with #{args[0].remote}"
        when :mkdir then
          # args[0] : local path name
          puts "creating directory #{args[0]}"
        when :finish then
          puts "all done!"
        end
      end
      puts "done!"
    else
      sftp.download! remote_file, local_file, :read_size => 16000
    end

    if options[:preserve]
      File.chmod attributes.permissions & 07777, local_file
      File.utime attributes.atime, attributes.mtime, local_file
    end
  end


  # Returns a fully-qualified host name (user@host:port) in the style of
  # the ssh command-line program for use in logging (and nothing more).
  # The ":port" syntax isn't valid SSH, but it's useful for logging purposes.
  def ssh_full_host(host, ssh_options)
    retval = ""
    retval << ssh_options[:user] << "@" if ssh_options[:user]
    retval << host
    retval << ":" << ssh_options[:port].to_s if ssh_options[:port]
    retval
  end

  def sleep_at_least(sleep_amount)
    while sleep_amount > 0
      sleep_amount -= sleep sleep_amount
    end
  end

end

end  # module ArkUtil
