#############################################################################
## Various methods useful for working with the topo-on-demand service.
##
## --------------------------------------------------------------------------
## Copyright (C) 2008 The Regents of the University of California.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
## $Id: todutils.rb,v 1.12 2017/06/16 01:22:03 youngh Exp $
#############################################################################

require 'arkutil/generalutils'
require 'json'

module ArkUtil

class TopoOnDemandUtils

  # This extracts the IP path from the textual trace line returned by the
  # topo-on-demand server.
  # 
  # This tries to behave like skitter_as_links for edge cases.  For
  # example, if there is a response from the destination, then this only
  # returns IP hops up to the hop position of the response from the
  # destination.  (Yes, strange enough, there can be hops past the hop
  # position of the destination.)  This includes all responses at the hop
  # position of the destination, even if the responses weren't from the
  # destination.
  #
  # This doesn't return the trailing gap.  So you can assume that if you
  # encounter a nil element before reaching the end of the result array,
  # then there will always be a non-nil element before reaching the end
  # of the array.
  #
  # This doesn't remove loops from the IP path, so you should do further
  # processing with find_all_loops() before generating, for example,
  # an AS path.
  #
  # --------------------------------------------------------------------------
  #
  # This returns an array with hop addresses at the corresponding array
  # positions (the source is at index 0).  If there is more than one
  # address at a given hop position, then this will use a subarray to
  # hold all the addresses.  Otherwise, this stores a single address
  # directly as a dotted-decimal string.  If there isn't a response at a
  # given hop position, then the corresponding array location will
  # contain a nil.
  def self.extract_ippath(trace)
    fields = trace.split "\t"
    return nil unless fields[0] == "T"  # XXX do more input validation

    retval = [ fields[1] ]

    dest_hop = nil
    if fields[6] == "R"
      response = fields[2] + "," + fields[7] + ",1"
      dest_hop = fields[8].to_i
      dest_index = 12 + dest_hop
      (fields.length .. dest_index - 1).each do |index|
        fields[index] = "q" unless fields[index]
      end

      case fields[dest_index]
      when "q", nil
        fields[dest_index] = response
      else
        fields[dest_index] += ";" + response
      end
    end

    fields[13..-1].each_with_index do |hop, index|
      index += 1  # adjust to take into account the source at index 0
      next if hop == "q"
      break if dest_hop && index > dest_hop

      hop.split(";").each do |response|
        address, rtt, attempt = response.split ","

        if retval[index]
          if retval[index].instance_of? Array
            retval[index] << address
          else
            retval[index] = [ retval[index], address ]
          end
        else
          retval[index] = address
        end
      end
    end

    retval
  end


  def self.extract_ippath_from_json(trace_json)
    begin
      trace = JSON.load trace_json
    rescue => e
      printf "extract_ippath_from_json: malformed JSON: %p\n", e
      return nil
    end

    retval = [ trace["src"] ]
    destination = trace["dst"]
    dest_responded = (trace["stop_reason"] == "NONE" ||
                      trace["stop_reason"] == "COMPLETED")
    dest_hop = 256

    # 'hops' can be missing
    (trace["hops"] || []).each do |hop|
      index, address = hop["probe_ttl"], hop["addr"]

      if retval[index]
        if retval[index].instance_of? Array
          retval[index] << address
        else
          retval[index] = [ retval[index], address ]
        end
      else
        retval[index] = address
      end

      if dest_responded && address == destination
        dest_hop = index if index < dest_hop
      end
    end

    # Clear all hops after the hop of the responding destination.
    retval.slice!(dest_hop + 1 .. -1) if dest_responded

    retval
  end


  # Returns an array of all loops found in the given IP path.
  #
  # You can apply this method to any trace, including traces without loops.
  # This returns an empty array if there are no loops.  Otherwise, each array
  # element specifies [ start_index, length, address ].
  #
  # The start index is the 0-based starting position of a loop.
  # A loop that appears in adjacent hops (e.g., 'B B' in 'A B B C') has
  # length 1.  A loop like 'B D B' has length 2.
  #
  # A path like 'A B B B' has two loops, both of length 1, starting at
  # indexes 1 and 2; that is, [1, 1, B] and [2, 1, B].
  #
  # Multiple responses at a hop don't affect the determination of loops in
  # a path.  That is, for the purposes of determining loops, it's as if there
  # were only one instance of each address at any given hop.
  def self.find_all_loops(ippath)
    retval = []  # [ [ start_index, length, address ] ]

    last_index = {}  # IP address => index of last occurrence of the address
    ippath.each_with_index do |addresses, index|
      next unless addresses

      ( addresses.instance_of?(Array) ? addresses.uniq : [ addresses ] ).
        each do |address|
        if last_index[address]
          length = index - last_index[address]
          retval << [ last_index[address], length, address ]
        end
        last_index[address] = index
      end
    end

    retval.sort  # mainly to order by starting index
  end

end


#============================================================================
#============================================================================

# A class for extracting links from traceroute IP paths, taking into account
# "anonymous" routers (that is, non-responding hops).
class TODTracerouteLinksExtractor

  attr_accessor :verbose

  def initialize
    @anonymous_id = 0

    # NODE => { NODE => frequency }}
    @links = Hash.new { |h,k| h[k] = Hash.new { |h2,k2| h2[k2] = 0 } }
    @anon_links = {}  # [ src, dest ] => anon ID
    @anon_predecessor = {}  # anon ID => src address
  end

  # {path} should be in the format output by TopoOnDemandUtils#extract_ippath.
  def add_path(path)
    @flattened_links = nil
    extract_all_links path
  end

  def links
    unless @flattened_links
      @flattened_links = []
      @links.each do |src, dest_hash|
        dest_hash.each do |dest, count|
          @flattened_links << [src, dest, count]
        end
      end
    end
    @flattened_links
  end

  def raw_links
    @links
  end

  def anon_predecessor(anon_id)
    @anon_predecessor[anon_id]
  end


  private #----------------------------------------------------------------

  def extract_all_links(path)
    return if path.length < 2

    # Handle the case of nil components at the beginning of the path.
    src_index = 0
    while src_index < path.length - 1 && path[src_index] == nil
      src_index += 1
    end

    dest_index = src_index + 1
    while src_index < path.length - 1
      while dest_index < path.length && path[dest_index] == nil
        dest_index += 1
      end
      return unless dest_index < path.length

      src = path[src_index]
      dest = path[dest_index]
      gap_length = dest_index - src_index - 1

      printf "index: src=%d, dest=%d; src=%p, dest=%p\n",
        src_index, dest_index, src, dest if @verbose

      extract_links src, dest, gap_length

      src_index = dest_index
      dest_index += 1
    end
  end


  # Extracts all possible links (that is, the cross product) from the
  # given pair of path components that are separated by {gap_length} gaps
  # (0 if the components are adjacent).  The path components, which must
  # not be nil (since nil's represent gaps), can be either a single value
  # or an array of values.
  def extract_links(src, dest, gap_length)
    simple_src = !src.instance_of?(Array)
    simple_dest = !dest.instance_of?(Array)

    if simple_src && simple_dest
      add_link src, dest, gap_length
    elsif simple_src
      dest.uniq.each do |dest|
        add_link src, dest, gap_length
      end
    elsif simple_dest
      src.uniq.each do |src|
        add_link src, dest, gap_length
      end
    else
      src.uniq.each do |src_value|
        dest.uniq.each do |dest_value|
          add_link src_value, dest_value, gap_length
        end
      end
    end
  end


  def add_link(src, dest, gap_length)
    case gap_length
    when 0
      add_direct_link src, dest

    when 1
      anon_id = @anon_links[ [src, dest] ]
      unless anon_id
        anon_id = new_id()
        @anon_links[ [src, dest] ] = anon_id
        @anon_predecessor[anon_id] = src
      end
      add_direct_link src, anon_id
      add_direct_link anon_id, dest

    else
      subpath = [src]
      gap_length.times do
        anon_id = new_id()
        subpath << anon_id
        @anon_predecessor[anon_id] = src
      end
      subpath << dest

      subpath.each_cons(2) do |s, d|
        add_direct_link s, d
      end
    end
  end


  def add_direct_link(src, dest)
    @links[src][dest] += 1 unless src == dest
  end


  def new_id
    @anonymous_id += 1
    "**" + @anonymous_id.to_s + "**"
  end

end

end  # module ArkUtil
